#include "FirebaseAuth.h"

class AuthStateChangeCounter : public firebase::auth::AuthStateListener {
public:
	AuthStateChangeCounter() : num_state_changes_(0) {}

	virtual void OnAuthStateChanged(Auth* auth) {  // NOLINT
		num_state_changes_++;
		LOG("OnAuthStateChanged User %p (state changes %d)",
			auth->current_user(), num_state_changes_);
	}

	void CompleteTest(const char* test_name, int expected_state_changes) {
		CompleteTest(test_name, expected_state_changes, expected_state_changes);
	}

	void CompleteTest(const char* test_name, int min_state_changes,
		int max_state_changes) {
		const bool success = min_state_changes <= num_state_changes_ &&
			num_state_changes_ <= max_state_changes;
		LOG("%sAuthStateListener called %d time%s on %s.",
			success ? "" : "ERROR: ", num_state_changes_,
			num_state_changes_ == 1 ? "" : "s", test_name);
		num_state_changes_ = 0;
	}

private:
	int num_state_changes_;
};

class IdTokenChangeCounter : public firebase::auth::IdTokenListener {
public:
	IdTokenChangeCounter() : num_token_changes_(0) {}

	virtual void OnIdTokenChanged(Auth* auth) {  // NOLINT
		num_token_changes_++;
		LOG("OnIdTokenChanged User %p (token changes %d)",
			auth->current_user(), num_token_changes_);
	}

	void CompleteTest(const char* test_name, int token_changes) {
		CompleteTest(test_name, token_changes, token_changes);
	}

	void CompleteTest(const char* test_name, int min_token_changes,
		int max_token_changes) {
		const bool success = min_token_changes <= num_token_changes_ &&
			num_token_changes_ <= max_token_changes;
		LOG("%sIdTokenListener called %d time%s on %s.",
			success ? "" : "ERROR: ", num_token_changes_,
			num_token_changes_ == 1 ? "" : "s", test_name);
		num_token_changes_ = 0;
	}

private:
	int num_token_changes_;
};


class PhoneListener : public PhoneAuthProvider::Listener {
public:
	PhoneListener()
		: num_calls_on_verification_complete_(0),
		num_calls_on_verification_failed_(0),
		num_calls_on_code_sent_(0),
		num_calls_on_code_auto_retrieval_time_out_(0) {}

	void OnVerificationCompleted(Credential /*credential*/) override {
		LOG("PhoneListener: successful automatic verification.");
		num_calls_on_verification_complete_++;
	}

	void OnVerificationFailed(const std::string& error) override {
		LOG("ERROR: PhoneListener verification failed with error, %s",
			error.c_str());
		num_calls_on_verification_failed_++;
	}

	void OnCodeSent(const std::string& verification_id,
		const PhoneAuthProvider::ForceResendingToken&
		force_resending_token) override {
		LOG("PhoneListener: code sent. verification_id=%s",
			verification_id.c_str());
		verification_id_ = verification_id;
		force_resending_token_ = force_resending_token;
		num_calls_on_code_sent_++;
	}

	void OnCodeAutoRetrievalTimeOut(const std::string& verification_id) override {
		LOG("PhoneListener: auto retrieval timeout. verification_id=%s",
			verification_id.c_str());
		verification_id_ = verification_id;
		num_calls_on_code_auto_retrieval_time_out_++;
	}

	const std::string& verification_id() const { return verification_id_; }
	const PhoneAuthProvider::ForceResendingToken& force_resending_token() const {
		return force_resending_token_;
	}
	int num_calls_on_verification_complete() const {
		return num_calls_on_verification_complete_;
	}
	int num_calls_on_verification_failed() const {
		return num_calls_on_verification_failed_;
	}
	int num_calls_on_code_sent() const { return num_calls_on_code_sent_; }
	int num_calls_on_code_auto_retrieval_time_out() const {
		return num_calls_on_code_auto_retrieval_time_out_;
	}

private:
	std::string verification_id_;
	PhoneAuthProvider::ForceResendingToken force_resending_token_;
	int num_calls_on_verification_complete_;
	int num_calls_on_verification_failed_;
	int num_calls_on_code_sent_;
	int num_calls_on_code_auto_retrieval_time_out_;
};

FirebaseAuth::FirebaseAuth()
{
	LOG("FirebaseAuth()");
}
FirebaseAuth::~FirebaseAuth()
{
	LOG("~FirebaseAuth()");
}

void FirebaseAuth::init()
{
	LOG("Initializing the Auth with Firebase API.");	
	
	LOG("Created the Firebase app %x.", static_cast<int>(reinterpret_cast<intptr_t>(firebase_app)));
	// Create the Auth class for that App.

	::firebase::ModuleInitializer initializer;
	initializer.Initialize(firebase_app, nullptr, [](::firebase::App* app, void*) {
		::firebase::InitResult init_result;
		Auth::GetAuth(app, &init_result);
		return init_result;
		});
	while (initializer.InitializeLastResult().status() !=
		firebase::kFutureStatusComplete) {
		if (FirebaseProcessEvents(1000)) return;  // exit if requested
	}

	if (initializer.InitializeLastResult().error() != 0) {
		LOG("Failed to initialize Auth: %s",
			initializer.InitializeLastResult().error_message());
		FirebaseProcessEvents(2000);
		return;
	}

	m_auth = Auth::GetAuth(firebase_app);

	LOG("Created the Auth %x class for the Firebase app.",
		static_cast<int>(reinterpret_cast<intptr_t>(m_auth)));
}

void FirebaseAuth::release()
{
	LOG("FirebaseAuth::release()");
	delete m_auth;
}

bool FirebaseAuth::signInWithEmailAndPassword(char* username, char* password)
{
	Future<User*> sign_in_future = m_auth->SignInWithEmailAndPassword(username, password);
	WaitForSignInFuture( sign_in_future, "Auth::SignInWithEmailAndPassword() existing email and password", kAuthErrorNone, m_auth);

	if (sign_in_future.error() == kAuthErrorNone)
	{
		return true;
	}
	return false;
}

bool FirebaseAuth::signOut()
{
	if (m_auth->current_user() == nullptr)
	{
		LOG("No user signed in at creation time.");
		return false;
	}
	else
	{
		m_auth->SignOut();
		// Wait for the sign out to complete.
		WaitForSignOut(m_auth);

		if (m_auth->current_user() != nullptr)
		{
			return false;
		}
	}
	return true;
}

bool FirebaseAuth::isPlayerAuthenticated()
{
	return (m_auth->current_user() != nullptr);
}

bool FirebaseAuth::registerWithEmailAndPassword(char* username, char* password)
{
	Future<User*> register_account = m_auth->CreateUserWithEmailAndPassword(username, password);
	FirebaseAuth::WaitForSignInFuture(register_account, "CreateUserWithEmailAndPassword() to create user", kAuthErrorNone, m_auth);
	return (register_account.error() == kAuthErrorNone);
}

void FirebaseAuth::testcase()
{
	LOG("FirebaseAdmob::testcase()");	

	// It's possible for current_user() to be non-null if the previous run
	// left us in a signed-in state.
	if (m_auth->current_user() == nullptr) {
		LOG("No user signed in at creation time.");
	}
	else {
		LOG(
			"Current user uid(%s) name(%s) already signed in, so signing them out.",
			m_auth->current_user()->uid().c_str(),
			m_auth->current_user()->display_name().c_str());
		m_auth->SignOut();
	}

	// --- Credential copy tests -------------------------------------------------
	{
		Credential email_cred =
			EmailAuthProvider::GetCredential(kCustomEmail, kCustomPassword);
		Credential facebook_cred =
			FacebookAuthProvider::GetCredential(kTestAccessTokenBad);

		// Test copy constructor.
		Credential cred_copy(email_cred);

		// Test assignment operator.
		cred_copy = facebook_cred;
		(void)cred_copy;
	}

	// --- Custom Profile tests --------------------------------------------------
	{
		if (kTestCustomEmail) {
			// Test Auth::SignInWithEmailAndPassword().
			// Sign in with email and password that have already been registered.
			Future<User*> sign_in_future =
				m_auth->SignInWithEmailAndPassword(kCustomEmail, kCustomPassword);
			WaitForSignInFuture(sign_in_future,
				"Auth::SignInWithEmailAndPassword() existing "
				"(custom) email and password",
				kAuthErrorNone, m_auth);
			// Test SignOut() after signed in with email and password.
			if (sign_in_future.status() == ::firebase::kFutureStatusComplete) {
				m_auth->SignOut();
				if (m_auth->current_user() != nullptr) {
					LOG(
						"ERROR: current_user() returning %x instead of nullptr after "
						"SignOut()",
						m_auth->current_user());
				}
			}
		}
	}

	// --- StateChange tests -----------------------------------------------------
	{
		AuthStateChangeCounter counter;
		IdTokenChangeCounter token_counter;

		// Test notification on registration.
		m_auth->AddAuthStateListener(&counter);
		m_auth->AddIdTokenListener(&token_counter);
		// Expect notification immediately after registration.
		counter.CompleteTest("registration", 1);
		token_counter.CompleteTest("registration", 1);

		// Test notification on SignOut(), when already signed-out.
		m_auth->SignOut();
		counter.CompleteTest("SignOut() when already signed-out", 0);
		token_counter.CompleteTest("SignOut() when already signed-out", 0);

		// Test notification on SignIn().
		Future<User*> sign_in_future = m_auth->SignInAnonymously();
		WaitForSignInFuture(sign_in_future, "Auth::SignInAnonymously()",
			kAuthErrorNone, m_auth);
		// Notified when the user is about to change and after the user has
		// changed.
		counter.CompleteTest("SignInAnonymously()", 1, 4);
		token_counter.CompleteTest("SignInAnonymously()", 1, 5);

		// Refresh the token.
		if (m_auth->current_user() != nullptr) {
			Future<std::string> token_future = m_auth->current_user()->GetToken(true);
			WaitForFuture(token_future, "GetToken()", kAuthErrorNone);
			counter.CompleteTest("GetToken()", 0);
			token_counter.CompleteTest("GetToken()", 1);
		}

		// Test notification on SignOut(), when signed-in.
		LOG("Current user %p", m_auth->current_user());  // DEBUG
		m_auth->SignOut();
		// Wait for the sign out to complete.
		WaitForSignOut(m_auth);
		counter.CompleteTest("SignOut()", 1);
		token_counter.CompleteTest("SignOut()", 1);
		LOG("Current user %p", m_auth->current_user());  // DEBUG

		m_auth->RemoveAuthStateListener(&counter);
		m_auth->RemoveIdTokenListener(&token_counter);
	}

	// --- Auth tests ------------------------------------------------------------
	{
		UserLogin user_login(m_auth);  // Generate a random name/password
		user_login.Register();
		if (!user_login.user()) {
			LOG("ERROR: Could not register new user.");
		}
		else {
			// Test Auth::SignInAnonymously().
			{
				Future<User*> sign_in_future = m_auth->SignInAnonymously();
				WaitForSignInFuture(sign_in_future, "Auth::SignInAnonymously()",
					kAuthErrorNone, m_auth);
				ExpectTrue("SignInAnonymouslyLastResult matches returned Future",
					sign_in_future == m_auth->SignInAnonymouslyLastResult());

				// Test SignOut() after signed in anonymously.
				if (sign_in_future.status() == ::firebase::kFutureStatusComplete) {
					m_auth->SignOut();
					if (m_auth->current_user() != nullptr) {
						LOG(
							"ERROR: current_user() returning %x instead of nullptr after "
							"SignOut()",
							m_auth->current_user());
					}
				}
			}

			// Test Auth::FetchProvidersForEmail().
			{
				Future<Auth::FetchProvidersResult> providers_future =
					m_auth->FetchProvidersForEmail(user_login.email());
				WaitForFuture(providers_future, "Auth::FetchProvidersForEmail()",
					kAuthErrorNone);
				ExpectTrue(
					"FetchProvidersForEmailLastResult matches returned Future",
					providers_future == m_auth->FetchProvidersForEmailLastResult());

				const Auth::FetchProvidersResult* pro = providers_future.result();
				if (pro) {
					LOG("  email %s, num providers %d", user_login.email(),
						pro->providers.size());
					for (auto it = pro->providers.begin(); it != pro->providers.end();
						++it) {
						LOG("    * %s", it->c_str());
					}
				}
			}

			// Test Auth::SignInWithEmailAndPassword().
			// Sign in with email and password that have already been registered.
			{
				Future<User*> sign_in_future = m_auth->SignInWithEmailAndPassword(
					user_login.email(), user_login.password());
				WaitForSignInFuture(
					sign_in_future,
					"Auth::SignInWithEmailAndPassword() existing email and password",
					kAuthErrorNone, m_auth);
				ExpectTrue(
					"SignInWithEmailAndPasswordLastResult matches returned Future",
					sign_in_future == m_auth->SignInWithEmailAndPasswordLastResult());

				// Test SignOut() after signed in with email and password.
				if (sign_in_future.status() == ::firebase::kFutureStatusComplete) {
					m_auth->SignOut();
					if (m_auth->current_user() != nullptr) {
						LOG(
							"ERROR: current_user() returning %x instead of nullptr after "
							"SignOut()",
							m_auth->current_user());
					}
				}
			}

			// Test User::UpdateUserProfile
			{
				Future<User*> sign_in_future = m_auth->SignInWithEmailAndPassword(
					user_login.email(), user_login.password());
				WaitForSignInFuture(
					sign_in_future,
					"Auth::SignInWithEmailAndPassword() existing email and password",
					kAuthErrorNone, m_auth);
				if (sign_in_future.error() == kAuthErrorNone) {
					User* user = *sign_in_future.result();
					const char* kDisplayName = "Hello World";
					const char* kPhotoUrl = "http://test.com/image.jpg";
					User::UserProfile user_profile;
					user_profile.display_name = kDisplayName;
					user_profile.photo_url = kPhotoUrl;
					Future<void> update_profile_future =
						user->UpdateUserProfile(user_profile);
					WaitForFuture(update_profile_future, "User::UpdateUserProfile",
						kAuthErrorNone);
					if (update_profile_future.error() == kAuthErrorNone) {
						ExpectStringsEqual("User::display_name", kDisplayName,
							user->display_name().c_str());
						ExpectStringsEqual("User::photo_url", kPhotoUrl,
							user->photo_url().c_str());
					}
				}
			}

			// Sign in anonymously, link an email credential, reauthenticate with the
			// credential, unlink the credential and finally sign out.
			{
				Future<User*> sign_in_anonymously_future = m_auth->SignInAnonymously();
				WaitForSignInFuture(sign_in_anonymously_future,
					"Auth::SignInAnonymously", kAuthErrorNone, m_auth);
				if (sign_in_anonymously_future.error() == kAuthErrorNone) {
					User* user = *sign_in_anonymously_future.result();
					std::string email = CreateNewEmail();
					Credential credential =
						EmailAuthProvider::GetCredential(email.c_str(), kTestPassword);
					// Link with an email / password credential.
					Future<SignInResult> link_future =
						user->LinkAndRetrieveDataWithCredential(credential);
					WaitForSignInFuture(link_future,
						"User::LinkAndRetrieveDataWithCredential",
						kAuthErrorNone, m_auth);
					if (link_future.error() == kAuthErrorNone) {
						LogSignInResult(*link_future.result());
						Future<SignInResult> reauth_future =
							user->ReauthenticateAndRetrieveData(credential);
						WaitForSignInFuture(reauth_future,
							"User::ReauthenticateAndRetrieveData",
							kAuthErrorNone, m_auth);
						if (reauth_future.error() == kAuthErrorNone) {
							LogSignInResult(*reauth_future.result());
						}
						// Unlink email / password from credential.
						Future<User*> unlink_future =
							user->Unlink(credential.provider().c_str());
						WaitForSignInFuture(unlink_future, "User::Unlink", kAuthErrorNone,
							m_auth);
					}
					m_auth->SignOut();
				}
			}

			// Sign in user with bad email. Should fail.
			{
				Future<User*> sign_in_future_bad_email =
					m_auth->SignInWithEmailAndPassword(kTestEmailBad, kTestPassword);
				WaitForSignInFuture(sign_in_future_bad_email,
					"Auth::SignInWithEmailAndPassword() bad email",
					::firebase::auth::kAuthErrorUserNotFound, m_auth);
			}

			// Sign in user with correct email but bad password. Should fail.
			{
				Future<User*> sign_in_future_bad_password =
					m_auth->SignInWithEmailAndPassword(user_login.email(),
						kTestPasswordBad);
				WaitForSignInFuture(sign_in_future_bad_password,
					"Auth::SignInWithEmailAndPassword() bad password",
					::firebase::auth::kAuthErrorWrongPassword, m_auth);
			}

			// Try to create with existing email. Should fail.
			{
				Future<User*> create_future_bad = m_auth->CreateUserWithEmailAndPassword(
					user_login.email(), user_login.password());
				WaitForSignInFuture(
					create_future_bad,
					"Auth::CreateUserWithEmailAndPassword() existing email",
					::firebase::auth::kAuthErrorEmailAlreadyInUse, m_auth);
				ExpectTrue(
					"CreateUserWithEmailAndPasswordLastResult matches returned Future",
					create_future_bad ==
					m_auth->CreateUserWithEmailAndPasswordLastResult());
			}

			// Test Auth::SignInWithCredential() using email&password.
			// Use existing email. Should succeed.
			{
				Credential email_cred_ok = EmailAuthProvider::GetCredential(
					user_login.email(), user_login.password());
				Future<User*> sign_in_cred_ok =
					m_auth->SignInWithCredential(email_cred_ok);
				WaitForSignInFuture(sign_in_cred_ok,
					"Auth::SignInWithCredential() existing email",
					kAuthErrorNone, m_auth);
				ExpectTrue("SignInWithCredentialLastResult matches returned Future",
					sign_in_cred_ok == m_auth->SignInWithCredentialLastResult());
			}

			// Test Auth::SignInAndRetrieveDataWithCredential using email & password.
			// Use existing email. Should succeed.
			{
				Credential email_cred = EmailAuthProvider::GetCredential(
					user_login.email(), user_login.password());
				Future<SignInResult> sign_in_future =
					m_auth->SignInAndRetrieveDataWithCredential(email_cred);
				WaitForSignInFuture(sign_in_future,
					"Auth::SignInAndRetrieveDataWithCredential "
					"existing email",
					kAuthErrorNone, m_auth);
				ExpectTrue(
					"SignInAndRetrieveDataWithCredentialLastResult matches "
					"returned Future",
					sign_in_future ==
					m_auth->SignInAndRetrieveDataWithCredentialLastResult());
				if (sign_in_future.error() == kAuthErrorNone) {
					const SignInResult* sign_in_result = sign_in_future.result();
					if (sign_in_result != nullptr && sign_in_result->user) {
						LOG("SignInAndRetrieveDataWithCredential");
						LogSignInResult(*sign_in_result);
					}
					else {
						LOG(
							"ERROR: SignInAndRetrieveDataWithCredential returned no "
							"result");
					}
				}
			}

			// Use bad Facebook credentials. Should fail.
			{
				Credential facebook_cred_bad =
					FacebookAuthProvider::GetCredential(kTestAccessTokenBad);
				Future<User*> facebook_bad =
					m_auth->SignInWithCredential(facebook_cred_bad);
				WaitForSignInFuture(
					facebook_bad,
					"Auth::SignInWithCredential() bad Facebook credentials",
					kAuthErrorInvalidProviderId, m_auth);
			}

			// Use bad GitHub credentials. Should fail.
			{
				Credential git_hub_cred_bad =
					GitHubAuthProvider::GetCredential(kTestAccessTokenBad);
				Future<User*> git_hub_bad =
					m_auth->SignInWithCredential(git_hub_cred_bad);
				WaitForSignInFuture(
					git_hub_bad, "Auth::SignInWithCredential() bad GitHub credentials",
					kAuthErrorInvalidProviderId, m_auth);
			}

			// Use bad Google credentials. Should fail.
			{
				Credential google_cred_bad = GoogleAuthProvider::GetCredential(
					kTestIdTokenBad, kTestAccessTokenBad);
				Future<User*> google_bad = m_auth->SignInWithCredential(google_cred_bad);
				WaitForSignInFuture(
					google_bad, "Auth::SignInWithCredential() bad Google credentials",
					kAuthErrorInvalidCredential, m_auth);
			}

			// Use bad Google credentials, missing an optional parameter. Should fail.
			{
				Credential google_cred_bad =
					GoogleAuthProvider::GetCredential(kTestIdTokenBad, nullptr);
				Future<User*> google_bad = m_auth->SignInWithCredential(google_cred_bad);
				WaitForSignInFuture(
					google_bad, "Auth::SignInWithCredential() bad Google credentials",
					kAuthErrorInvalidCredential, m_auth);
			}

#if defined(__ANDROID__)
			// Use bad Play Games (Android-only) credentials. Should fail.
			{
				Credential play_games_cred_bad =
					PlayGamesAuthProvider::GetCredential(kTestServerAuthCodeBad);
				Future<User*> play_games_bad =
					m_auth->SignInWithCredential(play_games_cred_bad);
				WaitForSignInFuture(
					play_games_bad,
					"Auth:SignInWithCredential() bad Play Games credentials",
					kAuthErrorInvalidCredential, m_auth);
			}
#endif  // defined(__ANDROID__)

#if TARGET_OS_IPHONE
			// Test Game Center status/login
			{
				// Check if the current user is authenticated to GameCenter
				bool is_authenticated = GameCenterAuthProvider::IsPlayerAuthenticated();
				if (!is_authenticated) {
					LOG("Not signed into Game Center, skipping test.");
				}
				else {
					LOG("Signed in, testing Game Center authentication.");

					// Get the Game Center credential from the device
					Future<Credential> game_center_credential_future =
						GameCenterAuthProvider::GetCredential();
					WaitForFuture(
						game_center_credential_future,
						"GameCenterAuthProvider::GetCredential()",
						kAuthErrorNone);

					const AuthError credential_error =
						static_cast<AuthError>(game_center_credential_future.error());

					// Only attempt to sign in if we were able to get a credential.
					if (credential_error == kAuthErrorNone) {
						const Credential* gc_credential_ptr =
							game_center_credential_future.result();

						if (gc_credential_ptr == nullptr) {
							LOG("Failed to retrieve Game Center credential.");
						}
						else {
							Future<User*> game_center_user =
								m_auth->SignInWithCredential(*gc_credential_ptr);
							WaitForFuture(game_center_user,
								"Auth::SignInWithCredential() test Game Center "
								"credential signin",
								kAuthErrorNone);
						}
					}
				}
			}
#endif  // TARGET_OS_IPHONE

			// Use bad Twitter credentials. Should fail.
			{
				Credential twitter_cred_bad = TwitterAuthProvider::GetCredential(
					kTestIdTokenBad, kTestAccessTokenBad);
				Future<User*> twitter_bad =
					m_auth->SignInWithCredential(twitter_cred_bad);
				WaitForSignInFuture(
					twitter_bad, "Auth::SignInWithCredential() bad Twitter credentials",
					kAuthErrorInvalidProviderId, m_auth);
			}

			// Use bad OAuth credentials. Should fail.
			{
				Credential oauth_cred_bad = OAuthProvider::GetCredential(
					kTestIdProviderIdBad, kTestIdTokenBad, kTestAccessTokenBad);
				Future<User*> oauth_bad = m_auth->SignInWithCredential(oauth_cred_bad);
				WaitForSignInFuture(
					oauth_bad, "Auth::SignInWithCredential() bad OAuth credentials",
					kAuthErrorInvalidProviderId, m_auth);
			}

			// Test Auth::SendPasswordResetEmail().
			// Use existing email. Should succeed.
			{
				Future<void> send_password_reset_ok =
					m_auth->SendPasswordResetEmail(user_login.email());
				WaitForFuture(send_password_reset_ok,
					"Auth::SendPasswordResetEmail() existing email",
					kAuthErrorNone);
				ExpectTrue(
					"SendPasswordResetEmailLastResult matches returned Future",
					send_password_reset_ok == m_auth->SendPasswordResetEmailLastResult());
			}

			// Use bad email. Should fail.
			{
				Future<void> send_password_reset_bad =
					m_auth->SendPasswordResetEmail(kTestEmailBad);
				WaitForFuture(send_password_reset_bad,
					"Auth::SendPasswordResetEmail() bad email",
					::firebase::auth::kAuthErrorUserNotFound);
			}
		}
	}
	// --- User tests ------------------------------------------------------------
	// Test anonymous user info strings.
	{
		Future<User*> anon_sign_in_for_user = m_auth->SignInAnonymously();
		WaitForSignInFuture(anon_sign_in_for_user,
			"Auth::SignInAnonymously() for User", kAuthErrorNone,
			m_auth);
		if (anon_sign_in_for_user.status() == ::firebase::kFutureStatusComplete) {
			User* anonymous_user = anon_sign_in_for_user.result()
				? *anon_sign_in_for_user.result()
				: nullptr;
			if (anonymous_user != nullptr) {
				LOG("Anonymous uid is %s", anonymous_user->uid().c_str());
				ExpectStringsEqual("Anonymous user email", "",
					anonymous_user->email().c_str());
				ExpectStringsEqual("Anonymous user display_name", "",
					anonymous_user->display_name().c_str());
				ExpectStringsEqual("Anonymous user photo_url", "",
					anonymous_user->photo_url().c_str());
				ExpectStringsEqual("Anonymous user provider_id", kFirebaseProviderId,
					anonymous_user->provider_id().c_str());
				ExpectTrue("Anonymous user is_anonymous()",
					anonymous_user->is_anonymous());
				ExpectFalse("Anonymous user is_email_verified()",
					anonymous_user->is_email_verified());
				ExpectTrue("Anonymous user metadata().last_sign_in_timestamp != 0",
					anonymous_user->metadata().last_sign_in_timestamp != 0);
				ExpectTrue("Anonymous user metadata().creation_timestamp != 0",
					anonymous_user->metadata().creation_timestamp != 0);

				// Test User::LinkWithCredential(), linking with email & password.
				const std::string newer_email = CreateNewEmail();
				Credential user_cred = EmailAuthProvider::GetCredential(
					newer_email.c_str(), kTestPassword);
				{
					Future<User*> link_future =
						anonymous_user->LinkWithCredential(user_cred);
					WaitForSignInFuture(link_future, "User::LinkWithCredential()",
						kAuthErrorNone, m_auth);
				}

				// Test User::LinkWithCredential(), linking with same email & password.
				{
					Future<User*> link_future =
						anonymous_user->LinkWithCredential(user_cred);
					WaitForSignInFuture(link_future, "User::LinkWithCredential() again",
						::firebase::auth::kAuthErrorProviderAlreadyLinked,
						m_auth);
				}

				// Test User::LinkWithCredential(), linking with bad credential.
				// Call should fail and Auth's current user should be maintained.
				{
					const User* pre_link_user = m_auth->current_user();
					ExpectTrue("Test precondition requires active user",
						pre_link_user != nullptr);

					Credential twitter_cred_bad = TwitterAuthProvider::GetCredential(
						kTestIdTokenBad, kTestAccessTokenBad);
					Future<User*> link_bad_future =
						anonymous_user->LinkWithCredential(twitter_cred_bad);
					WaitForFuture(link_bad_future,
						"User::LinkWithCredential() with bad credential",
						kAuthErrorInvalidProviderId);
					ExpectTrue("Linking maintains user",
						m_auth->current_user() == pre_link_user);
				}

				// Test Auth::SignInWithCredential(), signing in with bad credential.
				// Call should fail, and Auth's current user should be maintained.
				{
					const User* pre_signin_user = m_auth->current_user();
					ExpectTrue("Test precondition requires active user",
						pre_signin_user != nullptr);
					Credential twitter_cred_bad = TwitterAuthProvider::GetCredential(
						kTestIdTokenBad, kTestAccessTokenBad);
					Future<User*> signin_bad_future =
						m_auth->SignInWithCredential(twitter_cred_bad);
					WaitForFuture(signin_bad_future,
						"Auth::SignInWithCredential() with bad credential",
						kAuthErrorInvalidProviderId, m_auth);
					ExpectTrue("Failed sign in maintains user",
						m_auth->current_user() == pre_signin_user);
				}

				UserLogin user_login(m_auth);
				user_login.Register();

				if (!user_login.user()) {
					LOG("Error - Could not create new user.");
				}
				else {
					// Test email user info strings.
					Future<User*> email_sign_in_for_user =
						m_auth->SignInWithEmailAndPassword(user_login.email(),
							user_login.password());
					WaitForSignInFuture(email_sign_in_for_user,
						"Auth::SignInWithEmailAndPassword() for User",
						kAuthErrorNone, m_auth);
					User* email_user = email_sign_in_for_user.result()
						? *email_sign_in_for_user.result()
						: nullptr;
					if (email_user != nullptr) {
						LOG("Email uid is %s", email_user->uid().c_str());
						ExpectStringsEqual("Email user email", user_login.email(),
							email_user->email().c_str());
						ExpectStringsEqual("Email user display_name", "",
							email_user->display_name().c_str());
						ExpectStringsEqual("Email user photo_url", "",
							email_user->photo_url().c_str());
						ExpectStringsEqual("Email user provider_id", kFirebaseProviderId,
							email_user->provider_id().c_str());
						ExpectFalse("Email user is_anonymous()",
							email_user->is_anonymous());
						ExpectFalse("Email user is_email_verified()",
							email_user->is_email_verified());
						ExpectTrue("Email user metadata().last_sign_in_timestamp != 0",
							email_user->metadata().last_sign_in_timestamp != 0);
						ExpectTrue("Email user metadata().creation_timestamp  != 0",
							email_user->metadata().creation_timestamp != 0);

						// Test User::GetToken().
						// with force_refresh = false.
						Future<std::string> token_no_refresh = email_user->GetToken(false);
						WaitForFuture(token_no_refresh, "User::GetToken(false)",
							kAuthErrorNone);
						LOG("User::GetToken(false) = %s",
							token_no_refresh.result()
							? token_no_refresh.result()->c_str()
							: "");

						// with force_refresh = true.
						Future<std::string> token_force_refresh =
							email_user->GetToken(true);
						WaitForFuture(token_force_refresh, "User::GetToken(true)",
							kAuthErrorNone);
						LOG("User::GetToken(true) = %s",
							token_force_refresh.result()
							? token_force_refresh.result()->c_str()
							: "");

						// Test Reload().
						Future<void> reload_future = email_user->Reload();
						WaitForFuture(reload_future, "User::Reload()", kAuthErrorNone);

						// Test User::Unlink().
						Future<User*> unlink_future = email_user->Unlink("firebase");
						WaitForSignInFuture(unlink_future, "User::Unlink()",
							::firebase::auth::kAuthErrorNoSuchProvider,
							m_auth);

						// Sign in again if user is now invalid.
						if (m_auth->current_user() == nullptr) {
							Future<User*> email_sign_in_again =
								m_auth->SignInWithEmailAndPassword(user_login.email(),
									user_login.password());
							WaitForSignInFuture(email_sign_in_again,
								"Auth::SignInWithEmailAndPassword() again",
								kAuthErrorNone, m_auth);
							email_user = email_sign_in_again.result()
								? *email_sign_in_again.result()
								: nullptr;
						}
					}
					if (email_user != nullptr) {
						// Test User::provider_data().
						const std::vector<UserInfoInterface*>& provider_data =
							email_user->provider_data();
						LOG("User::provider_data() returned %d interface%s",
							provider_data.size(),
							provider_data.size() == 1 ? "" : "s");
						for (size_t i = 0; i < provider_data.size(); ++i) {
							const UserInfoInterface* user_info = provider_data[i];
							LOG(
								"    UID() = %s\n"
								"    Email() = %s\n"
								"    DisplayName() = %s\n"
								"    PhotoUrl() = %s\n"
								"    ProviderId() = %s",
								user_info->uid().c_str(), user_info->email().c_str(),
								user_info->display_name().c_str(),
								user_info->photo_url().c_str(),
								user_info->provider_id().c_str());
						}

						// Test User::UpdateEmail().
						const std::string newest_email = CreateNewEmail();
						Future<void> update_email_future =
							email_user->UpdateEmail(newest_email.c_str());
						WaitForFuture(update_email_future, "User::UpdateEmail()",
							kAuthErrorNone);

						// Test User::UpdatePassword().
						Future<void> update_password_future =
							email_user->UpdatePassword(kTestPasswordUpdated);
						WaitForFuture(update_password_future, "User::UpdatePassword()",
							kAuthErrorNone);

						// Test User::Reauthenticate().
						Credential email_cred_reauth = EmailAuthProvider::GetCredential(
							newest_email.c_str(), kTestPasswordUpdated);
						Future<void> reauthenticate_future =
							email_user->Reauthenticate(email_cred_reauth);
						WaitForFuture(reauthenticate_future, "User::Reauthenticate()",
							kAuthErrorNone);

						// Test User::SendEmailVerification().
						Future<void> send_email_verification_future =
							email_user->SendEmailVerification();
						WaitForFuture(send_email_verification_future,
							"User::SendEmailVerification()", kAuthErrorNone);
					}
				}
			}
		}

		// Test User::Delete().
		const std::string new_email_for_delete = CreateNewEmail();
		Future<User*> create_future_for_delete =
			m_auth->CreateUserWithEmailAndPassword(new_email_for_delete.c_str(),
				kTestPassword);
		WaitForSignInFuture(
			create_future_for_delete,
			"Auth::CreateUserWithEmailAndPassword() new email for delete",
			kAuthErrorNone, m_auth);
		User* email_user_for_delete = create_future_for_delete.result()
			? *create_future_for_delete.result()
			: nullptr;
		if (email_user_for_delete != nullptr) {
			Future<void> delete_future = email_user_for_delete->Delete();
			WaitForFuture(delete_future, "User::Delete()", kAuthErrorNone);
		}
	}
	{
		// We end with a login so that we can test if a second run will detect
		// that we're already logged-in.
		Future<User*> sign_in_future = m_auth->SignInAnonymously();
		WaitForSignInFuture(sign_in_future, "Auth::SignInAnonymously() at end",
			kAuthErrorNone, m_auth);

		LOG("Anonymous uid(%s)", m_auth->current_user()->uid().c_str());
	}

	LOG("Completed Auth tests.");
}

// Don't return until `future` is complete.
// Print a message for whether the result mathes our expectations.
// Returns true if the application should exit.
bool FirebaseAuth::WaitForFuture(const FutureBase& future, const char* fn,
	AuthError expected_error, bool log_error) {
	// Note if the future has not be started properly.
	if (future.status() == ::firebase::kFutureStatusInvalid) {
		LOG("ERROR: Future for %s is invalid", fn);
		return false;
	}

	// Wait for future to complete.
	LOG("  Calling %s...", fn);
	while (future.status() == ::firebase::kFutureStatusPending) {
		if (FirebaseProcessEvents(10000)) return true;
	}

	// Log error result.
	if (log_error) {
		const AuthError error = static_cast<AuthError>(future.error());
		if (error == expected_error) {
			const char* error_message = future.error_message();
			if (error_message) {
				LOG("%s completed as expected", fn);
			}
			else {
				LOG("%s completed as expected, error: %d '%s'", fn, error,
					error_message);
			}
		}
		else {
			LOG("ERROR: %s completed with error: %d, `%s`", fn, error,
				future.error_message());
		}
	}
	return false;
}


bool FirebaseAuth::WaitForSignInFuture(Future<User*> sign_in_future, const char* fn,
	AuthError expected_error, Auth* auth) {
	if (WaitForFuture(sign_in_future, fn, expected_error)) return true;

	const User* const* sign_in_user_ptr = sign_in_future.result();
	const User* sign_in_user =
		sign_in_user_ptr == nullptr ? nullptr : *sign_in_user_ptr;
	const User* auth_user = auth->current_user();

	if (expected_error == ::firebase::auth::kAuthErrorNone &&
		sign_in_user != auth_user) {
		LOG("ERROR: future's user (%x) and current_user (%x) don't match",
			static_cast<int>(reinterpret_cast<intptr_t>(sign_in_user)),
			static_cast<int>(reinterpret_cast<intptr_t>(auth_user)));
	}

	return false;
}

bool FirebaseAuth::WaitForSignInFuture(const Future<SignInResult>& sign_in_future,
	const char* fn, AuthError expected_error,
	Auth* auth) {
	if (WaitForFuture(sign_in_future, fn, expected_error)) return true;

	const SignInResult* sign_in_result = sign_in_future.result();
	const User* sign_in_user = sign_in_result ? sign_in_result->user : nullptr;
	const User* auth_user = auth->current_user();

	if (expected_error == ::firebase::auth::kAuthErrorNone &&
		sign_in_user != auth_user) {
		LOG("ERROR: future's user (%x) and current_user (%x) don't match",
			static_cast<int>(reinterpret_cast<intptr_t>(sign_in_user)),
			static_cast<int>(reinterpret_cast<intptr_t>(auth_user)));
	}

	return false;
}

// Wait for the current user to sign out.  Typically you should use the
// state listener to determine whether the user has signed out.
bool FirebaseAuth::WaitForSignOut(firebase::auth::Auth* auth) {
	while (auth->current_user() != nullptr) {
		if (FirebaseProcessEvents(1000)) return true;
	}
	// Wait - hopefully - long enough for listeners to be signalled.
	FirebaseProcessEvents(1000);
	return false;
}

// Create an email that will be different from previous runs.
// Useful for testing creating new accounts.
std::string FirebaseAuth::CreateNewEmail() {
	std::stringstream email;
	email << "random_" << std::time(0) << "@gmail.com";
	return email.str();
}

void FirebaseAuth::ExpectFalse(const char* test, bool value) {
	if (value) {
		LOG("ERROR: %s is true instead of false", test);
	}
	else {
		LOG("%s is false, as expected", test);
	}
}

void FirebaseAuth::ExpectTrue(const char* test, bool value) {
	if (value) {
		LOG("%s is true, as expected", test);
	}
	else {
		LOG("ERROR: %s is false instead of true", test);
	}
}

// Log results of a string comparison for `test`.
void FirebaseAuth::ExpectStringsEqual(const char* test, const char* expected,
	const char* actual) {
	if (strcmp(expected, actual) == 0) {
		LOG("%s is '%s' as expected", test, actual);
	}
	else {
		LOG("ERROR: %s is '%s' instead of '%s'", test, actual, expected);
	}
}

// Log a vector of variants.
void FirebaseAuth::LogVariantVector(const std::vector<Variant>& variants, int indent) {
	std::string indent_string(indent * 2, ' ');
	LOG("%s[", indent_string.c_str());
	for (auto it = variants.begin(); it != variants.end(); ++it) {
		const Variant& item = *it;
		if (item.is_fundamental_type()) {
			const Variant& string_value = item.AsString();
			LOG("%s  %s,", indent_string.c_str(), string_value.string_value());
		}
		else if (item.is_vector()) {
			LogVariantVector(item.vector(), indent + 2);
		}
		else if (item.is_map()) {
			LogVariantMap(item.map(), indent + 2);
		}
		else {
			LOG("%s  ERROR: unknown type %d", indent_string.c_str(),
				static_cast<int>(item.type()));
		}
	}
	LOG("%s]", indent_string.c_str());
}

// Log a map of variants.
void FirebaseAuth::LogVariantMap(const std::map<Variant, Variant>& variant_map,
	int indent) {
	std::string indent_string(indent * 2, ' ');
	for (auto it = variant_map.begin(); it != variant_map.end(); ++it) {
		const Variant& key_string = it->first.AsString();
		const Variant& value = it->second;
		if (value.is_fundamental_type()) {
			const Variant& string_value = value.AsString();
			LOG("%s%s: %s,", indent_string.c_str(), key_string.string_value(),
				string_value.string_value());
		}
		else {
			LOG("%s%s:", indent_string.c_str(), key_string.string_value());
			if (value.is_vector()) {
				LogVariantVector(value.vector(), indent + 1);
			}
			else if (value.is_map()) {
				LogVariantMap(value.map(), indent + 1);
			}
			else {
				LOG("%s  ERROR: unknown type %d", indent_string.c_str(),
					static_cast<int>(value.type()));
			}
		}
	}
}

// Display the sign-in result.
void FirebaseAuth::LogSignInResult(const SignInResult& result) {
	if (!result.user) {
		LOG("ERROR: User not signed in");
		return;
	}
	LOG("* User ID %s", result.user->uid().c_str());
	const AdditionalUserInfo& info = result.info;
	LOG("* Provider ID %s", info.provider_id.c_str());
	LOG("* User Name %s", info.user_name.c_str());
	LogVariantMap(info.profile, 0);
	const UserMetadata& metadata = result.meta;
	LOG("* Sign in timestamp %d",
		static_cast<int>(metadata.last_sign_in_timestamp));
	LOG("* Creation timestamp %d",
		static_cast<int>(metadata.creation_timestamp));
}
